from abc import ABC
from dataclasses import dataclass
from typing import Callable, Generic, TypeVar, Union

from .sentinels import Missing, MissingType

T = TypeVar("T")


class ParameterSpecification(ABC):
    pass


@dataclass(frozen=True)
class Param(Generic[T], ParameterSpecification):
    default: Union[T, MissingType] = Missing
    default_factory: Union[Callable[[], T], MissingType] = Missing

    def __post_init__(self):
        if self.default is not Missing and self.default_factory is not Missing:
            raise ValueError("cannot specify both `default` and `default_factory`")

    def get_default(self) -> Union[T, MissingType]:
        if self.default_factory is not Missing:
            return self.default_factory()
        elif self.default is not Missing:
            return self.default

        return Missing

    def has_default(self) -> bool:
        return self.default is not Missing or self.default_factory is not Missing



from .api import get_arguments, get_parameters, params
from .enums import ParameterType
from .errors import MissingSpecification, ResolutionError
from .manager import ParameterManager, ParamManager
from .models import Arguments, BoundArguments, Parameter
from .parameters import ParameterSpecification
from .resolvers import Resolver, Resolvers
from .sentinels import Missing
from .wrappers import Param



import inspect
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Generic, Tuple, TypeVar, Union

from .enums import ParameterType
from .parameters import ParameterSpecification
from .sentinels import Missing, MissingType
from .utils import parse

T = TypeVar("T")
S = TypeVar("S", bound=ParameterSpecification)


@dataclass(frozen=True)
class Arguments:
    args: Tuple[Any, ...] = field(default_factory=tuple)
    kwargs: Dict[str, Any] = field(default_factory=dict)

    def call(self, func: Callable[..., T], /) -> T:
        return func(*self.args, **self.kwargs)


@dataclass(frozen=True)
class BoundArguments:
    args: Dict[str, Any] = field(default_factory=dict)
    kwargs: Dict[str, Any] = field(default_factory=dict)

    def call(self, func: Callable[..., T], /) -> T:
        return func(*self.args.values(), **self.kwargs)

    @property
    def arguments(self) -> Dict[str, Any]:
        return {**self.args, **self.kwargs}


@dataclass(frozen=True)
class Parameter:
    name: str
    default: Union[Any, MissingType] = Missing
    annotation: Union[Any, MissingType] = Missing
    type: ParameterType = ParameterType.POSITIONAL_OR_KEYWORD

    @classmethod
    def from_parameter(cls, parameter: inspect.Parameter, /) -> "Parameter":
        return cls(
            name=parameter.name,
            default=parse(parameter.default),
            annotation=parse(parameter.annotation),
            type=ParameterType.from_kind(parameter.kind),
        )


@dataclass(frozen=True)
class Resolvable(Generic[S]):
    parameter: Parameter
    specification: S
    argument: Union[Any, MissingType] = Missing



import functools
import inspect
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Generic, Iterable, Optional, Type, TypeVar

from typing_extensions import ParamSpec

from .errors import ResolutionError
from .models import Arguments, BoundArguments, Parameter, Resolvable
from .parameters import Param, ParameterSpecification
from .resolvers import RESOLVERS, Resolver, Resolvers
from .sentinels import Missing

PS = ParamSpec("PS")
RT = TypeVar("RT")

R = TypeVar("R", bound=Callable)


def _bind_arguments(func: Callable, arguments: Arguments) -> BoundArguments:
    signature: inspect.Signature = inspect.signature(func)

    bound_arguments: inspect.BoundArguments = arguments.call(signature.bind)

    bound_arguments.apply_defaults()

    bound_args: Dict[str, Any] = dict(zip(signature.parameters, bound_arguments.args))
    bound_kwargs: Dict[str, Any] = bound_arguments.kwargs

    return BoundArguments(args=bound_args, kwargs=bound_kwargs)


@dataclass
class ParameterManager(Generic[R]):
    resolvers: Resolvers[R]

    @staticmethod
    def get_parameters(func: Callable, /) -> Dict[str, Parameter]:
        return {
            parameter.name: Parameter.from_parameter(parameter)
            for parameter in inspect.signature(func).parameters.values()
        }

    def get_parameter_specification(
        self, parameter: Parameter, /
    ) -> Optional[ParameterSpecification]:
        if isinstance(parameter.default, ParameterSpecification):
            return parameter.default
        else:
            return None

    def get_resolver(self, param_cls: Type[ParameterSpecification], /) -> R:
        resolver: Optional[R] = self.resolvers.get(param_cls)

        if resolver is not None:
            return resolver
        else:
            raise ResolutionError(f"No resolver for parameter {param_cls}")

    def resolve(self, resolvable: Resolvable) -> Any:
        raise ResolutionError("Resolution method not implemented")

    def resolve_all(
        self,
        resolvables: Iterable[Resolvable],
        /,
    ) -> Dict[str, Any]:
        return {
            resolvable.parameter.name: self.resolve(resolvable)
            for resolvable in resolvables
        }

    def get_resolvables(
        self, func: Callable, arguments: Arguments, /
    ) -> Dict[str, Resolvable]:
        bound_arguments: BoundArguments = _bind_arguments(func, arguments)
        parameters: Dict[str, Parameter] = self.get_parameters(func)
        resolvables: Dict[str, Resolvable] = {}

        parameter_name: str
        argument: Any
        for parameter_name, argument in bound_arguments.arguments.items():
            parameter: Parameter = parameters[parameter_name]
            specification: Optional[
                ParameterSpecification
            ] = self.get_parameter_specification(parameter)

            if specification is None:
                continue

            if argument is specification:
                argument = Missing

            resolvable: Resolvable = Resolvable(
                parameter=parameter, specification=specification, argument=argument
            )

            resolvables[parameter_name] = resolvable

        return resolvables

    def get_arguments(self, func: Callable, arguments: Arguments) -> BoundArguments:
        resolvables: Dict[str, Resolvable] = self.get_resolvables(func, arguments)
        parameters: Dict[str, Parameter] = self.get_parameters(func)
        bound_arguments: BoundArguments = _bind_arguments(func, arguments)
        resolved_arguments: Dict[str, Any] = self.resolve_all(resolvables.values())

        args: Dict[str, Any] = {}
        kwargs: Dict[str, Any] = {}

        parameter: Parameter
        for parameter in parameters.values():
            destination: Dict[str, Any]

            if parameter.name in bound_arguments.args:
                destination = args
            else:
                destination = kwargs

            argument: Any

            if parameter.name in resolvables:
                argument = resolved_arguments[parameter.name]
            else:
                argument = bound_arguments.arguments[parameter.name]

            destination[parameter.name] = argument

        return BoundArguments(args=args, kwargs=kwargs)

    def params(self, func: Callable[PS, RT], /) -> Callable[PS, RT]:
        @functools.wraps(func)
        def wrapper(*args: PS.args, **kwargs: PS.kwargs) -> RT:
            arguments: Arguments = Arguments(args=args, kwargs=kwargs)

            bound_arguments: BoundArguments = self.get_arguments(func, arguments)

            return bound_arguments.call(func)

        return wrapper


@dataclass
class ParamManager(ParameterManager[Resolver]):
    resolvers: Resolvers[Resolver] = field(default_factory=lambda: RESOLVERS)

    def get_parameter_specification(
        self, parameter: Parameter, /
    ) -> ParameterSpecification:
        specification: Optional[
            ParameterSpecification
        ] = super().get_parameter_specification(parameter)

        if specification is not None:
            return specification
        else:
            return Param(default=parameter.default)

    def resolve(
        self,
        resolvable: Resolvable,
    ) -> Any:
        resolver_cls: Type[ParameterSpecification] = type(resolvable.specification)
        resolver: Resolver = self.get_resolver(resolver_cls)

        return resolver(resolvable)



from enum import Enum, auto
from inspect import _ParameterKind


class NoValue(Enum):
    def __repr__(self) -> str:
        return f"<{type(self).__name__}.{self.name}>"


class ParameterType(NoValue):
    POSITIONAL_ONLY = auto()
    POSITIONAL_OR_KEYWORD = auto()
    VAR_POSITIONAL = auto()
    KEYWORD_ONLY = auto()
    VAR_KEYWORD = auto()

    @classmethod
    def from_kind(cls, kind: _ParameterKind, /) -> "ParameterType":
        return getattr(cls, kind.name)



from typing import Any, Callable, TypeVar, overload

T = TypeVar("T")

@overload
def Param() -> Any: ...
@overload
def Param(*, default: T) -> T: ...
@overload
def Param(*, default_factory: Callable[[], T]) -> T: ...



from typing import Any, Callable, Protocol, Type, TypeVar

from roster import Register

from .errors import ResolutionError
from .models import Resolvable
from .parameters import Param, ParameterSpecification
from .sentinels import Missing

R = TypeVar("R", bound=Callable)


class Resolver(Protocol):
    def __call__(self, resolvable: Resolvable, /) -> Any:
        ...


class Resolvers(Register[Type[ParameterSpecification], R]):
    pass


RESOLVERS: Resolvers[Resolver] = Resolvers()


@RESOLVERS(Param)
def resolve_param(resolvable: Resolvable[Param], /) -> Any:
    if resolvable.argument is not Missing:
        return resolvable.argument
    elif resolvable.specification.has_default():
        return resolvable.specification.get_default()
    else:
        raise ResolutionError("No value provided and parameter has no default")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['param']

package_data = \
{'': ['*']}

install_requires = \
['roster>=0.1.8,<0.2.0']

setup_kwargs = {
    'name': 'tombulled-param',
    'version': '0.1.8',
    'description': 'Enhanced function parameters',
    'long_description': '# param\nEnhanced function parameters\n\n## Installation\n```console\npip install git+https://github.com/tombulled/param.git@main\n```\n\n## Usage\n### Functions\nThe `@params` annotation works seamlessly with functions:\n```python\nfrom param import Param, params\n\n@params\ndef get(url: str, params: dict = Param(default_factory=dict)):\n    print("GET", url, params)\n```\n```python\n>>> get("https://httpbin.com/get")\nGET https://httpbin.com/get {}\n```\n\n### Classes\nThe `@params` annotation also works seamlessly with classes. Importantly, the `@params` annotation should be specified before `@staticmethod` or `@classmethod`.\n```python\nfrom param import Param, params\n\nclass MyClass:\n    @params\n    def get(self, url: str, params: dict = Param(default_factory=dict)):\n        print("GET", url, params)\n\n    @classmethod\n    @params\n    def post(cls, url: str, params: dict = Param(default_factory=dict)):\n        print("POST", url, params)\n\n    @staticmethod\n    @params\n    def put(url: str, params: dict = Param(default_factory=dict)):\n        print("PUT", url, params)\n```\n```python\n>>> obj = MyClass()\n>>>\n>>> obj.get("https://httpbin.com/get")\nGET https://httpbin.com/get {}\n>>>\n>>> obj.post("https://httpbin.com/post")\nPOST https://httpbin.com/post {}\n>>> MyClass.post("https://httpbin.com/post")\nPOST https://httpbin.com/post {}\n>>>\n>>> obj.put("https://httpbin.com/put")\nPUT https://httpbin.com/put {}\n>>> MyClass.put("https://httpbin.com/put")\nPUT https://httpbin.com/put {}\n```',
    'author': 'Tom Bulled',
    'author_email': '26026015+tombulled@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/param/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import inspect
from typing import Any, Union

from .sentinels import Missing, MissingType


def parse(value: Any, /) -> Union[Any, MissingType]:
    if value is inspect.Parameter.empty:
        return Missing
    else:
        return value



from typing import Callable, Dict, TypeVar

from typing_extensions import ParamSpec

from .manager import ParameterManager, ParamManager
from .models import Arguments, BoundArguments, Parameter

PS = ParamSpec("PS")
RT = TypeVar("RT")

MANAGER: ParameterManager = ParamManager()


def get_parameters(func: Callable, /) -> Dict[str, Parameter]:
    return MANAGER.get_parameters(func)


def get_arguments(func: Callable, arguments: Arguments) -> BoundArguments:
    return MANAGER.get_arguments(func, arguments)


def params(func: Callable[PS, RT], /) -> Callable[PS, RT]:
    return MANAGER.params(func)
